/* a simplest zone type: a list of IP addresses with
 * a common text (maybe with a substitution for IP).
 * A list is stored as a sorted array, with binary search --
 * maybe not the most efficient way, but it's the most optimal
 * from the memory requiriments point of view, simple to
 * handle and sufficient enouth (imho)
 */

#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "rbldnsd.h"
#include "dns.h"
#include "sstr.h"

struct zonedata {
  struct in_addr a;
  char *txt;
  unsigned naddr;
  struct in_addr *addr;
};

static struct zonedata *freezone(struct zonedata *z) {
  if (z) {
    if (z->txt) sstrfree(z->txt);
    if (z->addr) free(z->addr);
    free(z);
  }
  return NULL;
}

int ipset_query(const struct zonedata *const z, struct dnspacket *p,
            const unsigned char *const query, unsigned qflg)
{
   int a = 0, b = z->naddr - 1, m;
   struct in_addr addr;
   if (!dn4octet(query, &addr))
     return 0;
   while(a <= b) {
     m = (a + b) >> 1;
     if (z->addr[m].s_addr == addr.s_addr) {
       if (qflg & QUERY_A) addrec_a(p, z->a);
       if (qflg & QUERY_TXT) addrec_txt(p, z->txt, inet_ntoa(addr));
       return 1;
     }
     else if (z->addr[m].s_addr < addr.s_addr) a = m + 1;
     else b = m - 1;
   }
   return 0;
}

static int cmpaddr(const void *a, const void *b) {
  if (((struct in_addr*)a)->s_addr < ((struct in_addr*)b)->s_addr) return -1;
  if (((struct in_addr*)a)->s_addr > ((struct in_addr*)b)->s_addr) return 1;
  return 0;
}

struct zonedata *ipset_load(const char *name, struct zonedata *z) {
  int lineno = 0, l;
  struct in_addr *addr = NULL;
  unsigned naddr = 0, nalc = 0;
  struct in_addr a;
  char *p;
  FILE *f;
  char buf[128];
  freezone(z);
  if ((f = fopen(name, "r")) == NULL) {
    log(LOG_ERR, "unable to open zone file `%s': %s", name, strerror(errno));
    return NULL;
  }
  if ((z = (struct zonedata *)emalloc(sizeof(*z))) == NULL) {
    fclose(f);
    return NULL;
  }
  z->addr = NULL;
  z->txt = NULL;
  z->a = default_aaddr;
  while(fgets(buf, sizeof(buf), f)) {
    ++lineno;
    l = strlen(buf);
    if (l && buf[l-1] == '\n') buf[--l] = '\0';
    if (!l || buf[0] == '#') continue;
    if (lineno == 1 && buf[0] == ':') {
      if (!addrtxt(buf, &a, &p)) {
        log(LOG_WARNING, "zone %s line 1: invalid default entry", name);
	continue;
      }
      if (a.s_addr) z->a = a;
      if (*p && (z->txt = esstrdup(p)) == NULL) {
        fclose(f);
	return freezone(z);
      }
      continue;
    }
    for(p = buf; *p; ++p)
      if (*p == ' ' || *p == '\t' || *p == '#') break;
    *p = '\0'; /* skip any additional info */
    if ((l = ip4range(buf, &a)) == 0) {
      log(LOG_WARNING, "zone %s line %d: invalid entry", name, lineno);
      continue;
    }
    if (naddr + l >= nalc) {
      do nalc = nalc ? nalc * 2 : 512;
      while (naddr + l > nalc);
      z->addr = addr = (struct in_addr *)erealloc(addr, nalc * sizeof(*addr));
      if (!addr) {
	fclose(f);
	return freezone(z);
      }
    }
    while(l--)
      addr[naddr++].s_addr = a.s_addr++;
  }
  fclose(f);
  z->addr = (struct in_addr *)realloc(addr, naddr * sizeof(*addr));
  z->naddr = naddr;
  qsort(addr, naddr, sizeof(*addr), cmpaddr);
  return z;
}
